#include <time.h>
#include <utime.h>
#include "t3f/t3f.h"
#include "t3net/dlc.h"
#include "minizip/unzip.h"
#include "minizip/ioapi.h"
#include "init.h"

#define WRITEBUFFERSIZE (8192)
#ifdef __APPLE__
	// In darwin and perhaps other BSD variants off_t is a 64 bit value, hence no need for specific 64 bit functions
	#define FOPEN_FUNC(filename, mode) fopen(filename, mode)
	#define FTELLO_FUNC(stream) ftello(stream)
	#define FSEEKO_FUNC(stream, offset, origin) fseeko(stream, offset, origin)
#else
	FILE * fopen64(const char *filename, const char *type);
	#define FOPEN_FUNC(filename, mode) fopen64(filename, mode)
	#define FTELLO_FUNC(stream) ftello64(stream)
	#define FSEEKO_FUNC(stream, offset, origin) fseeko64(stream, offset, origin)
#endif

/* change_file_date : change the date/time of a file
    filename : the filename of the file where date/time must be modified
    dosdate : the new date at the MSDos format (4 bytes)
    tmu_date : the SAME new date at the tm_unz format */
void change_file_date(const char *filename, uLong dosdate, tm_unz tmu_date)
{
	struct utimbuf ut;
	struct tm newdate;
	newdate.tm_sec = tmu_date.tm_sec;
	newdate.tm_min=tmu_date.tm_min;
	newdate.tm_hour=tmu_date.tm_hour;
	newdate.tm_mday=tmu_date.tm_mday;
	newdate.tm_mon=tmu_date.tm_mon;
	if (tmu_date.tm_year > 1900)
		newdate.tm_year=tmu_date.tm_year - 1900;
	else
		newdate.tm_year=tmu_date.tm_year ;
	newdate.tm_isdst=-1;

	ut.actime=ut.modtime=mktime(&newdate);
	utime(filename,&ut);
}

/* extract current file from zip */
int do_extract_currentfile(unzFile uf, const char * destination)
{
    char filename_inzip[256];
    int err=UNZ_OK;
    FILE *fout=NULL;
    void* buf;
    uInt size_buf;
    ALLEGRO_PATH * out_path = NULL;
    ALLEGRO_PATH * inzip_path = NULL;
    ALLEGRO_PATH * sub_path = NULL;

//	printf("dest: %s\n", destination);
	out_path = al_create_path(destination);
	if(!out_path)
	{
		return UNZ_INTERNALERROR;
	}

    unz_file_info64 file_info;
	if(unzGetCurrentFileInfo64(uf, &file_info, filename_inzip, sizeof(filename_inzip), NULL, 0, NULL, 0) != UNZ_OK)
	{
		return UNZ_INTERNALERROR;
	}

    size_buf = WRITEBUFFERSIZE;
    buf = (void*)malloc(size_buf);
	if(buf==NULL)
	{
		printf("Error allocating memory\n");
		return UNZ_INTERNALERROR;
	}

    inzip_path = al_create_path(filename_inzip);
    if(!inzip_path)
    {
		return UNZ_INTERNALERROR;
	}
	sub_path = al_clone_path(out_path);
	if(!sub_path)
	{
		return UNZ_INTERNALERROR;
	}

	al_join_paths(sub_path, inzip_path);
	al_set_path_filename(sub_path, NULL);
	al_make_directory(al_path_cstr(sub_path, '/')); // recreate directory structure of ZIP into destination path
	al_join_paths(out_path, inzip_path);

        const char* write_filename;

        write_filename = al_path_cstr(out_path, '/');

        err = unzOpenCurrentFile(uf);
        if (err!=UNZ_OK)
        {
            printf("error %d with zipfile in unzOpenCurrentFile\n",err);
        }

        fout=FOPEN_FUNC(write_filename,"wb");
        if (fout!=NULL)
        {
//            printf(" extracting: %s\n",write_filename);

            do
            {
                err = unzReadCurrentFile(uf,buf,size_buf);
                if (err<0)
                {
                    printf("error %d with zipfile in unzReadCurrentFile\n",err);
                    break;
                }
                if (err>0)
                {
                    if (fwrite(buf,err,1,fout)!=1)
                    {
                        printf("error in writing extracted file\n");
                        err=UNZ_ERRNO;
                        break;
                    }
				}
            } while (err>0);

			if (fout)
			{
				fclose(fout);
			}
            if (err==0)
            {
                change_file_date(write_filename,file_info.dosDate, file_info.tmu_date);
			}
        }

        if (err==UNZ_OK)
        {
            err = unzCloseCurrentFile (uf);
            if (err!=UNZ_OK)
            {
                printf("error %d with zipfile in unzCloseCurrentFile\n",err);
            }
        }
        else
        {
            unzCloseCurrentFile(uf); /* don't lose the error */
		}

    free(buf);
    return err;
}

bool vgolf_unzip_zip(const char * zfn, const char * path)
{
    unzFile uf;
    unz_global_info gi;
    int i, x = 0;

	uf = unzOpen(zfn);
	if(uf == NULL)
	{
		printf("Failed to open ZIP file: %s\n", zfn);
		return false;
	}

	if(unzGetGlobalInfo(uf,&gi) != UNZ_OK)
	{
		printf("Failed to get ZIP file info\n");
		unzClose(uf);
		return false;
	}

	for(i = 0; i < gi.number_entry; i++)
	{
		if(do_extract_currentfile(uf, path) != UNZ_OK)
		{
			break;
		}
		else
		{
			x++;
		}
		if((i + 1) < gi.number_entry)
		{
			if(unzGoToNextFile(uf) != UNZ_OK)
			{
				break;
			}
		}
	}
	unzClose(uf);

	/* didn't successfully extract everything */
	if(x < gi.number_entry)
	{
		return false;
	}
	return true;
}

static int vgolf_dlc_download_tick = 0;
static char vgolf_dlc_download_description[128] = {0};

void vgolf_dlc_download_callback(void)
{
	char progress[128] = {0};
	int i;

	for(i = 0; i < (vgolf_dlc_download_tick / 4) % 3 + 1; i++)
	{
		strcat(progress, ".");
	}
	vgolf_show_load_screen(al_map_rgba_f(0.0, 0.0, 0.0, 0.75), vgolf_dlc_download_description, progress, "...");
	vgolf_dlc_download_tick++;
}

/* download DLC and extract into specified path */
bool vgolf_download_dlc(T3NET_DLC_LIST * lp, int item, const char * path)
{
	char buffer[1024];
	const char * zipfilename = t3f_get_filename(t3f_temp_path, "vgolf_dlc.zip", buffer, 1024);
	const ALLEGRO_FS_INTERFACE * old_fs;
	bool ret = false;

	t3net_set_download_callback(vgolf_dlc_download_callback);
	vgolf_dlc_download_tick = 0;
	sprintf(vgolf_dlc_download_description, "Downloading %s", lp->item[item]->name);
	old_fs = al_get_fs_interface();
	al_set_standard_fs_interface();
	if(t3net_download_file(lp->item[item]->url, zipfilename))
	{
		if(vgolf_unzip_zip(zipfilename, path))
		{
			lp->item[item]->flags = 1; // set flag to indicate DLC has been downloaded
			ret = true;
		}
	}
	al_set_fs_interface(old_fs);
	strcpy(vgolf_dlc_download_description, "");
	t3net_set_download_callback(NULL);
	return ret;
}
