#include "t3f/t3f.h"
#include "joynet/joynet.h"
#include <speex/speex.h>

#include "main.h"
#include "voip.h"

static ALLEGRO_AUDIO_RECORDER * vgolf_audio_recorder = NULL;
static bool vgolf_voip_initialized = false;

/* speex encode */
static void * vgolf_speex_encode_state = NULL;
static SpeexBits vgolf_speex_encode_bits;

/* speex decode */
static void * vgolf_speex_decode_state[VGOLF_VOIP_MAX_STREAMS] = {NULL};
static SpeexBits vgolf_speex_decode_bits[VGOLF_VOIP_MAX_STREAMS];
static ALLEGRO_AUDIO_STREAM * vgolf_voip_stream[VGOLF_VOIP_MAX_STREAMS] = {NULL};
static short * vgolf_voip_buffer[VGOLF_VOIP_MAX_STREAMS] = {NULL};
int vgolf_voip_buffer_read_pos[VGOLF_VOIP_MAX_STREAMS] = {0};
int vgolf_voip_buffer_write_pos[VGOLF_VOIP_MAX_STREAMS] = {0};
int vgolf_voip_buffer_size[VGOLF_VOIP_MAX_STREAMS] = {0};

/* debugging stuff */
static ALLEGRO_FILE * record_file = {NULL};

/* handle VOIP-related events */
void vgolf_handle_voip_event(ALLEGRO_EVENT * event)
{
	ALLEGRO_AUDIO_RECORDER_EVENT * revent;
	short * fragment_data;
	char encoded_bits[256];
	int bytes_written;
	int data_used = 0;
	int fragments;
	int i, j;
	ALLEGRO_AUDIO_STREAM * stream;
	
	if(vgolf_voip_initialized)
	{
		/* received audio input from microphone */
		if(event->type == ALLEGRO_EVENT_AUDIO_RECORDER_FRAGMENT)
		{
			revent = al_get_audio_recorder_event(event);
			fragment_data = revent->buffer;
			
			/* encode fragment into speex frames until all of fragment is used up */
			while(data_used < revent->samples)
			{
				/* encode the fragment */
				speex_bits_reset(&vgolf_speex_encode_bits);
				speex_encode_int(vgolf_speex_encode_state, fragment_data, &vgolf_speex_encode_bits);
				bytes_written = speex_bits_write(&vgolf_speex_encode_bits, encoded_bits, 256);
				
				/* send frame through JoyNet */
				joynet_send_client_voip_data(vgolf_client, encoded_bits, bytes_written);

				data_used += VGOLF_VOIP_FRAME_SIZE;
			}
		}
		
		/* audio stream wants to be filled */
		else if(event->type == ALLEGRO_EVENT_AUDIO_STREAM_FRAGMENT)
		{
			/* figure out which stream this fragment belongs to */
			stream = (ALLEGRO_AUDIO_STREAM *)event->any.source;
			for(i = 0; i < VGOLF_VOIP_MAX_STREAMS; i++)
			{
				if(vgolf_voip_stream[i] == stream)
				{
					break;
				}
			}
			
			if(i < VGOLF_VOIP_MAX_STREAMS)
			{
				/* use as many fragments as are available until we run out of data */
				fragments = al_get_available_audio_stream_fragments(stream);
				for(j = 0; j < fragments; j++)
				{
					fragment_data = al_get_audio_stream_fragment(vgolf_voip_stream[i]);
					if(fragment_data)
					{
						/* only fill if there is enough data */
						if(vgolf_voip_buffer_size[i] >= VGOLF_VOIP_BUFFER_SIZE)
						{
							/* put data from decode buffer into fragment */
							memcpy(fragment_data, &vgolf_voip_buffer[i][vgolf_voip_buffer_read_pos[i]], sizeof(short) * VGOLF_VOIP_BUFFER_SIZE);
							
							/* record stream to file for debugging */
							if(record_file)
							{
								al_fwrite(record_file, &vgolf_voip_buffer[i][vgolf_voip_buffer_read_pos[i]], sizeof(short) * VGOLF_VOIP_BUFFER_SIZE);
							}
							
							/* advance decode ring buffer */
							vgolf_voip_buffer_read_pos[i] += VGOLF_VOIP_BUFFER_SIZE;
							if(vgolf_voip_buffer_read_pos[i] >= VGOLF_VOIP_DECODE_BUFFER_SIZE)
							{
								vgolf_voip_buffer_read_pos[i] -= VGOLF_VOIP_DECODE_BUFFER_SIZE;
							}
							vgolf_voip_buffer_size[i] -= VGOLF_VOIP_BUFFER_SIZE;
						}
						/* give the fragment back to Allegro */
						al_set_audio_stream_fragment(stream, fragment_data);
					 }
				}
			}
		}
	}
	t3f_event_handler(event); // send event through T3F's event handler
}

/* initialize VOIP system */
bool vgolf_initialize_voip(void)
{
	int setting;
	
	/* create audio recorder so we can get input from microphone */
	vgolf_audio_recorder = al_create_audio_recorder(1000, VGOLF_VOIP_FRAME_SIZE, VGOLF_VOIP_RATE, ALLEGRO_AUDIO_DEPTH_INT16, ALLEGRO_CHANNEL_CONF_1);
	if(vgolf_audio_recorder)
	{
		/* set up speex encoder */
		vgolf_speex_encode_state = speex_encoder_init(&speex_nb_mode);
		setting = 8;
		speex_encoder_ctl(vgolf_speex_encode_state, SPEEX_SET_QUALITY, &setting);
		speex_bits_init(&vgolf_speex_encode_bits);
		record_file = al_fopen("test.raw", "wb");
		
		/* start the audio recorder for microphone input */
		al_register_event_source(t3f_queue, al_get_audio_recorder_event_source(vgolf_audio_recorder));
		al_start_audio_recorder(vgolf_audio_recorder);

		vgolf_voip_initialized = true;
		
		return true;
	}
	return false;
}

/* shut down the VOIP system */
void vgolf_shut_down_voip(void)
{
	int i;

	if(vgolf_voip_initialized)
	{
		/* close debug stuff */
		if(record_file)
		{
			al_fclose(record_file);
			record_file = NULL;
		}
		
		/* stop the audio recorder */
		al_stop_audio_recorder(vgolf_audio_recorder);
		
		/* free speex encoder data */
		speex_encoder_destroy(vgolf_speex_encode_state);
		speex_bits_destroy(&vgolf_speex_encode_bits);
		
		/* destroy the audio recorder */
		al_destroy_audio_recorder(vgolf_audio_recorder);
		vgolf_audio_recorder = NULL;
		
		/* destroy any VOIP streams and related data created by the VOIP system during usage */
		for(i = 0; i < VGOLF_VOIP_MAX_STREAMS; i++)
		{
			if(vgolf_voip_stream[i])
			{
				/* free speex decoder data */
				speex_decoder_destroy(vgolf_speex_decode_state[i]);
				speex_bits_destroy(&vgolf_speex_decode_bits[i]);
				
				/* free audio stream */
				al_destroy_audio_stream(vgolf_voip_stream[i]);
				vgolf_voip_stream[i] = NULL;

				/* free decode buffer */
				al_free(vgolf_voip_buffer[i]);
				vgolf_voip_buffer[i] = NULL;
			}
		}
	}
}

/* when client receives voip data, this function will decode and buffer it,
 * when the stream is ready for a fragment it will be filled from the buffer if 
 * there is enough data in the buffer */
int vgolf_voip_callback(int client, void * data, int data_size)
{
	int setting;
	
//	printf("voip callback\n");
	
	/* create stream on first use */
	if(vgolf_voip_initialized)
	{
		if(!vgolf_voip_stream[client])
		{
			/* create decode buffer */
			vgolf_voip_buffer[client] = al_malloc(sizeof(short) * VGOLF_VOIP_DECODE_BUFFER_SIZE);
			if(!vgolf_voip_buffer[client])
			{
				goto error_clean;
			}
			vgolf_voip_buffer_read_pos[client] = 0;
			vgolf_voip_buffer_write_pos[client] = 0;
			vgolf_voip_buffer_size[client] = 0;
				
			/* create voip audio stream for playback */
			vgolf_voip_stream[client] = al_create_audio_stream(4, VGOLF_VOIP_BUFFER_SIZE, VGOLF_VOIP_RATE, ALLEGRO_AUDIO_DEPTH_INT16, ALLEGRO_CHANNEL_CONF_1);
			if(!vgolf_voip_stream[client])
			{
				goto error_clean;
			}
			al_register_event_source(t3f_queue, al_get_audio_stream_event_source(vgolf_voip_stream[client]));
			al_attach_audio_stream_to_mixer(vgolf_voip_stream[client], al_get_default_mixer());

			/* set up speex decoder */
			vgolf_speex_decode_state[client] = speex_decoder_init(&speex_nb_mode);
			if(!vgolf_speex_decode_state[client])
			{
				goto error_clean;
			}
			setting = 1;
			speex_decoder_ctl(vgolf_speex_decode_state[client], SPEEX_SET_ENH, &setting);
			speex_bits_init(&vgolf_speex_decode_bits[client]);
		}
		
		/* decode the audio */
//		printf("break 1\n");
		speex_bits_read_from(&vgolf_speex_decode_bits[client], data, data_size);
//		printf("break 2\n");
//		printf("write pos = %d (%lu, %lu)\n", vgolf_voip_buffer_write_pos[client], vgolf_voip_buffer[client], &(vgolf_voip_buffer[client][vgolf_voip_buffer_write_pos[client]]));
		speex_decode_int(vgolf_speex_decode_state[client], &vgolf_speex_decode_bits[client], &(vgolf_voip_buffer[client][vgolf_voip_buffer_write_pos[client]]));
		vgolf_voip_buffer_write_pos[client] += VGOLF_VOIP_FRAME_SIZE;
		if(vgolf_voip_buffer_write_pos[client] >= VGOLF_VOIP_DECODE_BUFFER_SIZE)
		{
			vgolf_voip_buffer_write_pos[client] = 0;
		}
		vgolf_voip_buffer_size[client] += VGOLF_VOIP_FRAME_SIZE;
	}
//	printf("voip callback done\n");
	return 1;
	
	/* clean up on error */
	error_clean:
		printf("callback error\n");
		if(vgolf_voip_buffer[client])
		{
			al_free(vgolf_voip_buffer[client]);
			vgolf_voip_buffer[client] = NULL;
		}
		if(vgolf_voip_stream[client])
		{
			al_destroy_audio_stream(vgolf_voip_stream[client]);
			vgolf_voip_stream[client] = NULL;
		}
		if(vgolf_speex_decode_state[client])
		{
			speex_decoder_destroy(vgolf_speex_decode_state[client]);
			speex_bits_destroy(&vgolf_speex_decode_bits[client]);
		}
		return 0;
}
